/*
  VS1003b 16-Button Player
  ========================
  This source code package is provided as-is and without warranty.
  (C)2007-2009 VLSI Solution Oy

  This code is licensed for free use with any product using VS1003b.

  Because of the limited code size, almost all functions have been
  hand-optimized. The original C-code is included in this file, but
  is excluded from compilation using preprocessor directives.
  There may be small functional differences between the asm and C
  versions.

  Some features can be turned on and off with preprocessor definitions
  in Makefile and in this file. Features that are enabled have an
  effect on the code size, and the resulting program may not fit in
  the available code space. In this case the linker will give an error.

  SD-card mini player
  -------------------
  Software for the SD Card Mini Player is also compiled from this
  source using LINEIN_BUTTONS == 3.
*/

/*
  Note on MOSI/DREQ:

  DREQ is changed in normal firmware both in SDI interrupts when the
  stream buffer becomes too full, and whenever the stream buffer becomes
  empty enough.

  In this version SDI interrupt is not enabled, so it does not touch DREQ.
  So, we can use DREQ ourselves as we wish and access to it does not need to
  be protected by interrupt Disable()/Enable(). However, the normal decoding
  will still keep raising DREQ when it sees that there is space for new data.

 */

/*
  Start volume can be changed in the boot image or by modifying the
  VOL_VAL define in Makefile.

  See LOUDNESS_ON define in the Makefile if you want to set/clear
  loudness at power-on.
 */

#define USE_HC        /*SDHC support 54 words*/
/*Note: some EEPROM's can't handle high speed, and you must use a slower
        saveword function, see the alternate version in asm-saveword.s. */
//#define SAVE_POSITION /*Starts last played song after power-on, 3+29 words*/
//#define SAVE_VOLUME /*31+29 words */

#if 0 && LINEIN_BUTTONS == 16 && !defined(NO_WMA)
//#undef SAVE_POSITION
//#undef SAVE_VOLUME
//#define SONG_BUTTONS /*no volume or other buttons, just select */
#define USE_PLAY_MODES
#define PAUSE_BEFORE_TO_AFTER /*changes pause at poweron to pause after play*/
#define SHORT_AND_LONG_SELECT_SONG
#define LONG_PRESS_TIME 24 //48 /*48=1.5seconds*/
#define NO_RANDOM /* save words.. */
#endif

//#define NO_RANDOM /* no random play,  11+9=20 words */
//#define KEYTIMER /*send 0x00 to UART for 3 secs when button pressed. 21*/

//#define SHUFFLE_PLAY /*+17 words compared to random*/

#ifndef LONG_PRESS_TIME
#define LONG_PRESS_TIME 24 /*24=0.75seconds*/
#endif

#ifndef LINEIN_BUTTONS
#define LINEIN_BUTTONS 16
#endif

#if LINEIN_BUTTONS == 16 && !defined(NO_WMA)
#define SAVE_VOLUME
#endif



#include <string.h>
#include "standalone.h"

auto void play_frame(void);

/*
  mode A:when a song playing is over, player goes to pause mode.(default mode)
  mode B:loop song mode, the playing file is repeated until a button pushed.
  mode C:normal mode, the files are played one after another.
  When power on the player,
  VS1003 is goes to pause mode and waiting for press the keys.

  MODE A: pause before play (default)
  MODE B: loop song
  MODE C: normal mode

  Also set from c.s .uword 0x0004 -- SCI_AICTRL3 pause before play
  to enable pause mode at power-on.
 */
//#define MODE_ABC
//#define MODE_ABC_OWN_KEYS
#if 1
#define MODE_DIR //switch between directories (files 0..7 and 8..15)
#define SHORT_AND_LONG_SELECT_SONG
#endif
/*
  1 - toggle dir (file 0..7 and 8..15)
  2,3 - volume of left chn
  4,5 - volume of right chn
  6-  - select file #-6
 */

#if defined(MODE_ABC) //|| defined(MODE_DIR)
#define NO_RANDOM
#define USE_PLAY_MODES /*loop, pause after, pause before*/
#endif

#include <stdlib.h>

/*
  LINEIN_BUTTONS 16

  GND -----10k-+-10k-+-10k-+-..-+-10k-+-10k--- GBUF
               |     |     |    |     |
	       /     /     /    /     /
  GND ---1nF---+-----+-----+----+-----+---> LINE
  with headphones:
  b =  [-20350 -18850 -17450 -16100 -14750 -13500 -12260 -11000 -9800 -8500 -7300 -6200 -5000 -3750 -2500 -1250 0];
  plot(16+floor((b+1250/2)./1250),'-x');
  plot(floor((20000+b+1250/2)./1250),'-x');

  with amplifier:
  b = [-20272 -18725 -17350 -16025 -14640 -13420 -12180 -10920 -9750 -8500 -7360 -6180 -4950 -3780 -2470 -1270 0];

  MODU_DIV 64
  b = [-20300 -18875 -17500 -16125 -14825 -13550 -12275 -11050 -9850 -8625 -7425 -6200 -5000 -3775 -2525 -1275 100];

  LINEIN_BUTTONS 32

  GND -----10k-+-10k-+-10k-+-..-+-10k-+-10k--- GBUF
               |     |     |    |     |
	       /     /     /    /     /
  GND ---1nF---+-----+-----+----+-----+---> LINE
  with headphones:
  b = [-20880 -19950 -19100 -18290 -17475 -16730 -15980 -15280 -14575 -13890          -13260 -12600 -11980 -11380 -10760 -10170 -9540  -8940  -8390  -7800           -7240  -6640  -6100  -5500  -4950  -4340  -3780  -3150  -2600  -1940           -1320  -710    0];
  with amp:
  b = [-20880 -19930 -19115 -18225 -17490 -16700 -15990 -15280 -14550 -13900          -13200 -12615 -11950 -11360 -10725 -10175 -9540  -8900  -8370  -7810           -7150  -6660  -6100  -5425  -4950  -4325  -3760  -3180  -2580  -1965           -1255  -710    0];
  plot(floor((b-(4800+800/2))./-800),'-x');grid on;
  plot(floor((b-(600/2))./-600),'-x');grid on;

  amp, MODU_DIV 16
  b = [-20250 -18900 -17650 -16580 -15560 -14700 -13850 -13100 -12400 -11700          -11100 -10450 -9900  -9400  -8900  -8400  -7850  -7350  -6900  -6500           -6050  -5600  -5150  -4675  -4250  -3800  -3300  -2850  -2325  -1800           -1225  -650   +100];
  MODU_DIV 32
  b = [-20700 -19650 -18700 -17800 -16925 -16100 -15350 -14600 -13850 -13225          -12550 -11950 -11350 -10750 -10150 -9600  -9000  -8450  -7900  -7375           -6850  -6300  -5750  -5250  -4700  -4150  -3600  -3050  -2500  -1875           -1275  -650   +150];
  plot(t,floor((-100.*b+47.*b.*b+73.*b.^3+1)/2),'rx-');
  plot(t,floor((-106.*b+71.*b.*b+125.*b.^3+1)/2),'rx-');

  MODU_DIV 64
  b = [-20925 -20050 -19250 -18450 -17650 -16950 -16200 -15500 -14850 -14175          -13500 -12850 -12200 -11550 -10950 -10350 -9700  -9100  -8550  -7925           -7350  -6750  -6175  -5575  -4975  -4400  -3800  -3200  -2550  -1950           -1300  -650   +100];
  dump.m

  MODU_DIV 128  noise increases..



  MODU_DIV 32
  b = [-50 -720 -1340 -2000 -2500 -3100 -3680 -4230 -4770 -5310 -5845 -6380           -6910 -7440 -7975 -8515 -9065];


  3-button
  b = [0 -8192 -16384 -24576];

 */

//#define TIMEOUT /* timeout when decoding seems to be stuck, 28 words. */
//#define SKIP_ID3V2    /* skip ID3v2 tags (44 words) */
#define CHECKFILETYPE /* check file type according to short-filename suffix (9+22) */
#if !defined(CHECKFILETYPE) && !defined(MODE_ABC) && LINEIN_BUTTONS != 3
#define HDATCHECK 800 /* give up on a file after 800/2=400kB ~ 3-6sec (16 words) */
#endif/*MODE_ABC*/
#define KEEP_DRIVERS_OFF /* .. until MMC located (3 words) */
#define WITH_SUBDIRECTORIES /* support subdirectories, 36 words */
#define USE_LED          /* use LED to indicate loudness/random play (11 words) */

//#define LOOP_FILES     /* play each file until file changed */
//#define PAUSE_BEFORE_PLAY /*goto pause mode before each file*/
//#define PAUSE_AFTER_PLAY /*goto pause mode after each file, press play to play the same file again, press next to play the next file.*/
/* NO_WMA and NO_MIDI must be set in Makefile so that
   the suffixes are also excluded. */

void putch(register __a0 int ch); /* send a char to TX */


#ifdef CHECKFILETYPE
#undef CHECKFILETYPE /*temporary undef*/
#ifdef NO_WMA
#define CHECKFILETYPE CheckFileType1003Nw
#else
#define CHECKFILETYPE CheckFileType1003
#endif
#endif

#define CTRL3_NEXT_DIR          (1<<9) /* OpenFile() ignores ..*/

#define CTRL3_AT_END            (1<<6)
#define CTRL3_NO_NUMFILES       (1<<5)
#define CTRL3_PAUSE_ON          (1<<4)
#define CTRL3_FILE_READY        (1<<3)

#define CTRL3_PLAY_MODE_MASK    (3<<1)
#define CTRL3_PAUSE_AFTER_PLAY  (3<<1)
#define CTRL3_PAUSE_BEFORE_PLAY (2<<1)
#define CTRL3_LOOP_SONG         (1<<1)
#define CTRL3_PLAY_NEXT         (0<<1)

#define CTRL3_RANDOM_PLAY       (1<<0) /* SCI_AICTRL3 */

#ifndef GPIO_IDATA
#define GPIO_IDATA GPIO_DATA
#endif

#define SPI_xCS   1 /* GPIO */
#define SPI_CLK   8
#define SPI_MISO  4
#define SPI_MISO_SHIFT -3
#define SPI_xCS2  2
#define SPI_MOSI  1 /* DREQ */

#if defined(SAVE_POSITION) || defined(SAVE_VOLUME)
#define SPI_WREN  0x06
#define SPI_WRDI  0x04
#define SPI_RDSR  0x05
#define SPI_WRSR  0x01
#define SPI_READ  0x03
#define SPI_WRITE 0x02

/* Note: should use slower clock for SPI-EEPROM access! See c.s . */
register __a1 short SpiDelay(void);
auto u_int16 SpiSendReceiveSpi(register __a0 u_int16 datTopAligned, register __a1 s_int16 bits);
void SaveWord(register __c0 u_int16 addr, register __c1 u_int16 value);
#endif/*SAVE_POSITION*/
#ifdef SAVE_VOLUME
int volChanged = 0;
#endif


typedef enum {
    spi_pause=0,
    spi_file,
    spi_seek,
    spi_waitdata,
    spi_data,
    spi_zeros,
    spi_record,
    spi_waitima
} SPISTATE;

SPISTATE spiState, oldSpiState;
u_int16 spiCnt;
#ifdef TIMEOUT
u_int16 timeOut;
#endif

struct SPIX { /* do not touch! ASM uses the structure */
    s_int32 size;
    u_int32 sector;
    u_int16 fragSize; /* TODO: make this 32-bit and multiply fragment list clusters by sectorsPerCluster. */
    __y struct FRAGMENT *fragment;
} spix;


#define 	MMC_GO_IDLE_STATE   0
#define 	MMC_SEND_OP_COND   1
#define 	MMC_SEND_IF_COND   8
#define 	MMC_SEND_CSD   9
#define 	MMC_SEND_CID   10
#define 	MMC_SEND_STATUS   13
#define 	MMC_SET_BLOCKLEN   16
#define 	MMC_READ_SINGLE_BLOCK   17
#define 	MMC_WRITE_BLOCK   24
#define 	MMC_PROGRAM_CSD   27
#define 	MMC_SET_WRITE_PROT   28
#define 	MMC_CLR_WRITE_PROT   29
#define 	MMC_SEND_WRITE_PROT   30
#define 	MMC_TAG_SECTOR_START   32
#define 	MMC_TAG_SECTOR_END   33
#define 	MMC_UNTAG_SECTOR   34
#define 	MMC_TAG_ERASE_GROUP_START   35
#define 	MMC_TAG_ERARE_GROUP_END   36
#define 	MMC_UNTAG_ERASE_GROUP   37
#define 	MMC_ERASE   38
#define         MMC_READ_OCR     58
#define 	MMC_CRC_ON_OFF   59
#define 	MMC_R1_BUSY   0x80
#define 	MMC_R1_PARAMETER   0x40
#define 	MMC_R1_ADDRESS   0x20
#define 	MMC_R1_ERASE_SEQ   0x10
#define 	MMC_R1_COM_CRC   0x08
#define 	MMC_R1_ILLEGAL_COM   0x04
#define 	MMC_R1_ERASE_RESET   0x02
#define 	MMC_R1_IDLE_STATE   0x01
#define 	MMC_STARTBLOCK_READ   0xFE
#define 	MMC_STARTBLOCK_WRITE   0xFE
#define 	MMC_STARTBLOCK_MWRITE   0xFC
#define 	MMC_STOPTRAN_WRITE   0xFD
#define 	MMC_DE_MASK   0x1F
#define 	MMC_DE_ERROR   0x01
#define 	MMC_DE_CC_ERROR   0x02
#define 	MMC_DE_ECC_FAIL   0x04
#define 	MMC_DE_OUT_OF_RANGE   0x04
#define 	MMC_DE_CARD_LOCKED   0x04
#define 	MMC_DR_MASK   0x1F
#define 	MMC_DR_ACCEPT   0x05
#define 	MMC_DR_REJECT_CRC   0x0B
#define 	MMC_DR_REJECT_WRITE_ERROR   0x0D

#if 1
auto u_int16 MmcCommand(register __b0 s_int16 cmd, register __d u_int32 arg);
#else
auto u_int16 MmcCommand(register __b0 s_int16 cmd, register __d u_int32 arg) {
    register __b1 u_int16 t;
#if 1
    puts("Cmd(");
    puthex(cmd);
    puthex(arg>>16);
    puthex(arg);
    puts(")");
#endif
    /* some MMC brands require the extra clock pulses     */
    SpiSendClocks();
    SpiSendClocks();
    USEX(GPIO_ODATA) &= ~SPI_xCS2;

    while (SpiSendReceiveMmc(0xff00,8) != 0xff) {
	/* Unexpected busy signal from MMC */
    }
    SpiSendReceiveMmc(cmd, 16); /* top half normally 0xff */
    SpiSendReceiveMmc(arg>>16, 16);
    SpiSendReceiveMmc(arg, 16);
    t = SpiSendReceiveMmc(0x9500, 8);    /* Valid CRC for init, then don't care */

    cmd = 2000;               /* timeout value, try max. 100 times */
    while (cmd && (t & 0x80)) { /* R1 response not detected */
	t = SpiSendReceiveMmc(0xff00, 8);
	cmd--;               /* decrease the timeout value */
    }
    return t;
}
#endif

#if 1
auto u_int16 ReadDiskSector(register __a u_int32 sector);
#else
auto u_int16 ReadDiskSector(register __a u_int32 sector) {
    register s_int16 i;
    register u_int16 *buffer = stream_buffer;

#ifdef USE_HC /* In ASM this uses self-modified code */
    MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,(sector&0x7fffffffUL)<<hcShift);
#else
    MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,(sector<<9));
#endif
    do {
	i = SpiSendReceiveMmc(0xff00,8);
    } while (i == 0xff);

    if (i != 0xfe) {
	SpiSendClocks();/*USEX(GPIO_ODATA) |= SPI_xCS2;*/
	return 5;
    }

    for (i=0; i<512; i+=2) {
	*buffer++ = SpiSendReceiveMmc(0xffff,16);
    }
    SpiSendReceiveMmc(0xffff,16); /* discard crc */

    /* generate some extra SPI clock edges to finish up the command */
    SpiSendClocks();
    SpiSendClocks();
    return 0; /* All OK return */
}
#endif



#if 1
auto u_int16 GetByte(register __c0 u_int16 n);
#else
auto u_int16 GetByte(register __c0 u_int16 n) {
    register u_int16 *p = stream_buffer[n>>1];
    if (n & 1) {
	n = *p;
    } else {
	n = *p >> 8;
    }
    return n & 0xff;
}
#endif
#if 1
auto u_int16 GetWord(register __c0 u_int16 n);
#else
auto u_int16 GetWord(register __c0 u_int16 n) {
    return GetByte(n) | (GetByte(n+1)<<8);
}
#endif
#if 1
auto u_int32 GetLong(register __c0 u_int16 n);
#else
auto u_int32 GetLong(register __c0 u_int16 n) {
    return GetWord(n) | ((u_int32)GetWord(n+2)<<16);
}
#endif

#ifdef FATINFO_IN_Y
__y struct FATINFO fatInfo;
#else
struct FATINFO fatInfo;
#endif

#if 1
auto u_int16 InitFileSystem(void);
#else
auto u_int16 InitFileSystem(void) {
    /* Load MBR -- the first sector on disk */
    ReadDiskSector(0);

#if 0
    /* Ok, it should be a MBR sector. Let's verify */
    if (stream_buffer[255] != 0x55aa) {
	return -1; /* sector 0 is not MBR. */
    }
#endif

#if 0
    /* This checks that partition 1 is active. Alter code to allow
       other partition configurations. */
    if (GetByte(0x1be) == 0x80) {
    } else {
    }
#endif

    /* Now leave MBR and load sector 0 of partition */
    /* check for boot sector instead of partition table */
    if (GetLong(0x36)!= (((long)'1'<<24)|((long)'T'<<16)|((long)'A'<<8)|'F') &&
	GetLong(0x52)!= (((long)'3'<<24)|((long)'T'<<16)|((long)'A'<<8)|'F')
	) {
	fatInfo.currentSector = GetLong(0x1c6);
	ReadDiskSector(fatInfo.currentSector);
    }

    {
	u_int16 BPB_BlksPerSec;

	/* Determine FAT Type (16/32) */
	/* This should be done better, but it'll do for now. */
	fatInfo.BPB_RootEntCnt = GetWord(17);
	/* FAT32 does not have separate root entries. */
	/* TODO: should detect FAT32 by BPB_FATSz16==0 */
	fatInfo.IS_FAT_32 = fatInfo.BPB_RootEntCnt ? 0 : 1;
	fatInfo.FilSysType = GetWord(54+3); /* 0x3231 for FAT12 */
#if 0
	if (!IS_FAT_32){
	    if (GetByte(54+4)=='2') {/*diskSect.fat.ext._16.BS_FilSysType[4]=='2'*/) {
		fatInfo.IS_FAT_12 = GetByte(54);
	    }
	}
#endif

	/* OK, let's calculate */
	/* First, let's get rid of the idea that we have byte addresses
	   in the file system. Nope, let's only deal in physical disk
	   sectors of 512 bytes. First we convert the FAT byter per sector
	   value to "512B disk sectors per fat sector" value. */

	BPB_BlksPerSec = GetWord(11) >> 9;

	/* Then we adjust the Sector per Cluster to mean physical disk
	   sectors. in 99% of the cases it is already so because bytes
	   per sector almost always is 512 in FAT. Maximum cluster size
	   is 65536 bytes (128 disk sectors). */

	fatInfo.fatSectorsPerCluster = GetByte(13)/*BPB_SecPerClus*/ * BPB_BlksPerSec;
	fatInfo.fatStart = fatInfo.currentSector +
	    (u_int32)GetWord(14)/*BPB_RsvdSecCnt*/ * BPB_BlksPerSec;
	fatInfo.rootStart = GetWord(22);/*BPB_FATSz16*/
	if (fatInfo.rootStart == 0) {
	    if (!fatInfo.IS_FAT_32) 
		return 0x0b; /* should be FAT32; can not find root directory */
	    fatInfo.rootStart = GetLong(36);/*BPB_FATSz32*/
	}
	/* BPB_NumFATs * BPB_BlksPerSec fits in a word (8+7 bits) */
	fatInfo.rootStart = fatInfo.rootStart * (GetByte(16)/*BPB_NumFATs*/ *
						 BPB_BlksPerSec)
	    + fatInfo.fatStart;
	fatInfo.dataStart = fatInfo.rootStart + (fatInfo.BPB_RootEntCnt >> 4)
	    - (fatInfo.fatSectorsPerCluster << 1); /*first cluster is cluster 2*/
    }
    return 0;
}
#endif

#if 0
/** FAT/VFAT directory record union */
union DirRecordUnion {
  /** Standard directory entry */
  struct Entry {
    byte Name[11]; 0
    byte Attr;     11
    byte NTRes;    12
    byte CrtTimeTenth; 13
    word CrtTime;  14
    word CrtDate;  16
    word LstAccDate; 18
    word FstClusHi; 20
    word WrtTime;   22
    word WrtDate;   24
    word FstClusLo; 26
    u_32 FileSize;  28
  } entry;

  /** Extended directory entry */
  struct LongEntry {
    byte Ord;      0
    word Name1[5]; 1 /**< characters 1-5 */
    byte Attr;     11
    byte Type;     12 /**< entry type, zero=long name component */
    byte Chksum;   13
    word Name2[6]; 14 /**< characters 6-11 */
    word FstClusLO; 26 /**< zero for long entry */
    word Name3[2]; 28 /**< characters 12-13 */
  } longentry;
};
#endif


#define LAST_FRAGMENT 0x80000000UL
struct FRAGMENT {
    u_int32 start; /*high bit set if last fragment*/
    u_int16 size;
};
//#define MAX_FRAGMENTS 28//16//22
__y struct FRAGMENT fragments[MAX_FRAGMENTS];

#if 1
auto __y struct FRAGMENT *FragmentList(register __i2 __y struct FRAGMENT *frag,
			  register __b u_int32 fatCluster);
#else
auto __y struct FRAGMENT *FragmentList(register __i2 __y struct FRAGMENT *frag,
			  register __b u_int32 fatCluster) {
    register __d0 u_int16 offset;

    /* fatCluster means now current cluster, starting from 0, not 2. */
    /* note that startcluster=2 is already compensated in fatInitGlobals */
#if 0
    frag->size = 0;
    frag->start = fatCluster * fatInfo.fatSectorsPerCluster + fatInfo.dataStart;
#endif
    goto l55; /* read first sector */

    while (1) {
	u_int32 t;
	if (fatInfo.IS_FAT_32) {
	    t = GetLong(offset) & 0x0fffffffUL;
	} else {
	    t = GetWord(offset);
	    if (t == 0xffffU)
		t = 0x0fffffffUL;
	}
	frag->size += fatInfo.fatSectorsPerCluster;
	++fatCluster;
	if (frag->size >= 32768U || t != fatCluster) {
	    /* if not sequential cluster, start next fragment */
	    fatCluster = t;
	    if (frag == &fragments[MAX_FRAGMENTS-1] ||
		t == 0x0fffffffUL) {
		frag->start |= LAST_FRAGMENT;
		return frag+1;
	    }
	    frag++;
	l55:
	    frag->size = 0;
	    frag->start =
		fatCluster * fatInfo.fatSectorsPerCluster + fatInfo.dataStart;

	} else {
	    offset += (fatInfo.IS_FAT_32 ? 4 : 2);
	    offset &= 511;
	    if (offset != 0)
		continue;
	}
read:
	/* If FAT page is not already in memory, load it. */
	if (fatInfo.IS_FAT_32) {
	    offset = ((u_int16)fatCluster & 0x7F) << 2;
	    ReadDiskSector((fatCluster >> 7) + fatInfo.fatStart);
	} else {
	    offset = ((u_int16)fatCluster & 0xff) << 1;
	    ReadDiskSector((fatCluster >> 8) + fatInfo.fatStart);
	}
    }
}
#endif

u_int16 gFileNum[2];

#ifdef __VSDSP__
auto s_int16 OpenFile(register __c0 u_int16 fileNum);
#else
auto s_int16 OpenFile(register __c0 u_int16 fileNum) {
    register __i1 __y struct FRAGMENT *curFragment = &fragments[0];
    register __i0 __y struct FRAGMENT * __y nextFragment;

    /* Start at the start of root directory. */

    if (fatInfo.IS_FAT_32) {
	nextFragment = FragmentList(curFragment, 2);
    } else {
	curFragment->start = fatInfo.rootStart | LAST_FRAGMENT;
	curFragment->size  = (fatInfo.BPB_RootEntCnt >> 4); /* /sectorsPerCluster?*/
	nextFragment = curFragment + 1;
    }
    gFileNum[0] = fileNum;
    gFileNum[1] = 0;
    return HandleDir(curFragment, nextFragment);
}
#endif

#ifdef CHECKFILETYPE
#if 0
#define MKID(a,b,c) ((a)|((b)<<8)|((u_int32)(c)<<16))
auto s_int16 CheckFileType(register __a u_int32 name) {
    name &= 0x00ffffffUL;
    if (name == MKID('W','A','V')) {
	return -1;
    }
    if (name == MKID('M','P','3')) {
	return -1;
    }
    if (name == MKID('M','I','D')) {
	return -1;
    }
    if (name == MKID('W','M','A')) {
	return -1;
    }
    if (name == MKID('A','S','F')) {
	return -1;
    }
    return 0;
}
#endif/*1*/
#endif/*CHECKFILETYPE*/

auto s_int16 HandleDir(register __i2 __y struct FRAGMENT *curFragment,
		       __y struct FRAGMENT *nextFragment) {
    register __c1 int i;
    register u_int32 currentSector;

    currentSector = curFragment->start /*& 0x7fffffffUL */;
    i = 0;
    goto first;
    while (1) {
	register __d1 u_int16 fn = GetByte(i+0); /* first char of filename */

	/* We are now looking at FAT directory structure. */
	/* Is current file a regular file? */

	if (fn == 0)
	    goto eod; /*end of directory*/
	if (fn != 0xe5) {
	    u_int16 attrib = GetByte(i+11/*Attr*/);
#ifdef WITH_SUBDIRECTORIES
	    /* Is it a subdirectory? */
	    if ((attrib & 0x10) && fn != '.') {
		/* TODO: do not call if FAT12! */
		if (HandleDir(nextFragment,
			      FragmentList(nextFragment,
					   ((u_int32)GetWord(i+20)<<16)+GetWord(i+26))) < 0)
		    return -1;
		ReadDiskSector(currentSector);/*reread sector*/
	    } else /* it was a subdirectory */
#endif
	    /* Attributes: NO directory, NO volume id, NO system, NO hidden */
	    if ((attrib & 0xde) == 0
#ifdef CHECKFILETYPE
		/*
		  check supported files from name suffix:
		  vs1003:         wav mp3 mid wma asf
		*/
		&& CHECKFILETYPE(GetLong(i+8))
#endif
		) {

		/* It is a regular file. */

		{ /* force gFileNum[1]++ to a different register */
		    register u_int16 t = gFileNum[1] + 1;
		    gFileNum[1] = t;
		}
		if (gFileNum[0]-- == 0) {
		    /* ------------ FILE FOUND ------------- */
		    fatInfo.fileSize = GetLong(i+28);
		    FragmentList(&fragments[0],
				 ((u_int32)GetWord(i+20)<<16) + GetWord(i+26));
		    return -1; /* File found, All OK return */
		}
	    } /* normal file */
	} /* 0xe5 / 0x00 deleted or unused entry */
	i = (i + 32) & 511;
	if (i == 0) {
	    currentSector++;
	    if (--(curFragment->size) == 0) {
		if ((s_int16)(curFragment->start>>16) < 0) {
	eod:
		    return gFileNum[1];
		}
		curFragment++;
		currentSector = curFragment->start /*& 0x7fffffffUL*/;
	    }
	first:
	    ReadDiskSector(currentSector);
	}
    }
}


#ifdef HDATCHECK
u_int16 hdatCheck;
#endif
extern __y u_int16 hwSampleRate;
u_int16 __x * __y stream_save;

__y struct {
    s_int16 Data; /* must be first entry! Used by c.s */
    u_int16 Cnt;
    u_int16 Strobe;
    u_int16 Timer; /* Data,Cnt,Strobe,Timer must be in this order (buts.s) */
    u_int16 Key;
    s_int16 KeyTime;
#ifdef KEYTIMER
    s_int16 KeyTimer;
#endif
} ui;

#if 1
#ifdef ENABLE_UI
s_int16 ApplAddr(register __i0 s_int16 **d, register __a1 s_int16 mode,
		 register __a0 s_int16 n);
#endif
#else
#ifdef ENABLE_UI
s_int16 ApplAddr(register __i0 s_int16 **d, register __a1 s_int16 mode,
		 register __a0 s_int16 n) {
    if (mode == APPL_AUDIO) {
	register u_int16 t = hwSampleRate/16;
	/* generate a timebase for User Interface */
	ui.Cnt += n;
	if (ui.Cnt > t) {
	    ui.Cnt -= t;
	    ui.Timer++;
	    ui.Strobe = 1;
	}
    }
}
#endif/*ENABLE_UI*/
#endif


#ifdef USE_HC
s_int16 hcShift;
#endif

#if LINEIN_BUTTONS == 3
void Off(void) {
    SpiSendClocks();
#if 0 //def SAVE_POSITION
    SaveWord(28/*AICTRL0*/, USEX(SCI_AICTRL0));
#endif
    USEX(SCI_STATUS) |= (1<<SCIST_APDOWN2); /*drivers off*/
    USEX(SER_DREQ) = 0; /*LED off*/
#ifdef ENABLE_WATCHDOG
    do {
	USEX(WDOG_RESET) = WDOG_RESET_VAL;
    } while ((s_int16)USEX(MODU_DATA) < -20480);
#endif
    USEX(SCI_STATUS) |= (1<<SCIST_APDOWN1)|(1<<SCIST_APDOWN2);
    while (1)
	;
}
#endif

//s_int16 keyData;
/*
  User interface and data from MMC/SD to stream buffer happens here.
 */
void UserHook(void) {
    register __c1 SPISTATE cachedSpiState = spiState;

    if (ui.Strobe) { /* activates 16 times per second */
	ui.Strobe = 0;

#if 0
	{
	    register u_int16 t = USEX(MODU_DATA);
	    putch(t>>8);
	    putch(t);
	}
	ui.Data = 0;
#else
	/*
	  This part of code converts the ADC result into a key number.
	  0 means no key is pressed.
	 */

	{
	    register s_int16 t = USEX(MODU_DATA);
#if LINEIN_BUTTONS == 3
	    t = (t - 4096) / -8192;
	    if (t < 0)
		t = 0;
	    ui.Data = t;
#elif LINEIN_BUTTONS == 16
	    ui.Data = (t - (1250/2)) / -1250;
	    if (ui.Data < 0)
		ui.Data = 0;
#elif LINEIN_BUTTONS == 32
	    //MODU_DIV 32
	    ui.Data = (s_int16)
		((f_int16)256+
		 ( (f_int16)(-106*256) +
		   ( (f_int16)(71*256) + (f_int16)(125*256) * (f_int16)t) *
		   (f_int16)t) * (f_int16)t) / 512;
	    if (ui.Data < 0)
		ui.Data = 0;
#else
	    err('');
#endif
	}
#endif

#ifdef KEYTIMER
	if (ui.Data) {
	    ui.KeyTimer = 3*32;
	} else if (ui.KeyTimer > 0) {
	    ui.KeyTimer--;
	}
#endif
	/*
	  This code converts the raw keypress information into events.
	  Positive event values are short presses (shorter than 0.75 seconds),
	  negative event values are long presses (longer than 0.75 seconds).
	  Long press events will continue 16 times (after the start time)
	  per second as long as the key is being pressed.
	  One-shot long press can be implemented by comparing the time
	  or by setting keyTime to -32767.
	 */
	{
	    register int event = 0;
	    if (ui.Data == ui.Key) {
		if (ui.KeyTime < 4*32) {
		    ui.KeyTime++;
		}
		if (/*ui.Key &&*/ ui.KeyTime >= LONG_PRESS_TIME) {
		    /* long press event (or not if ui.Key==0) */
		    event = -ui.Key;
		}
	    } else {
		if (ui.KeyTime > 1) {
		    /* the same value read at least three times
		       -- to prevent noise causing button presses
		       when no keys have been pressed. */
		    if (ui.KeyTime < LONG_PRESS_TIME) {
			/* was pressed < 0.75 seconds */
			event = ui.Key; /* short press event */
		    }
		}
		ui.Key = ui.Data;
		ui.KeyTime = 0;
	    }

#if LINEIN_BUTTONS == 3

#if 0 /*custom/normal*/
	    /*event 3 is the power/pause/play button*/
	    if (event == -1) {  /*volume down -- long press*/
		register u_int16 vol = USEX(SCI_VOL);
		vol += 0x0101;
	    setvol:
		USEX(SCI_VOL) = vol;
		SetVolumeFromSpi();
#ifdef SAVE_VOLUME
		volChanged = 1;
#endif/*SAVE_VOLUME*/
	    } else if (event == -2) {   /*volume up -- long press*/
		register u_int16 vol = USEX(SCI_VOL);
		if (vol) {
		    vol -= 0x0101;
		    goto setvol;
		}
	    } else if (abs(event) == 3) {
		/* short or long press of SW1(?) -- power off */
		Off();
	    } else if (event > 0) {
		/* short press of next/prev buttons */
		register u_int16 song = USEX(SCI_AICTRL0);
		if (event == 2) {
		    song += 1; /* next file */
		} else if (USEX(SCI_DECODE_TIME) < 5) {
		    if (song == 0) {
			song = USEX(SCI_AICTRL1)-1;
		    } else { /* previous file */
			song -= 1;
		    }
		    /* else the same file */
		}
		USEX(SCI_AICTRL0) = song;
		USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
		cachedSpiState = spi_zeros;
	    }
#else/*custom/normal*/
	    if (event == -1) {  /*volume down -- long press*/
		register u_int16 vol = USEX(SCI_VOL);
		vol += 0x0101;
	    setvol:
		USEX(SCI_VOL) = vol;
		SetVolumeFromSpi();
#ifdef SAVE_VOLUME
		volChanged = 1;
#endif/*SAVE_VOLUME*/
	    } else if (event == -2) {   /*volume up -- long press*/
		register u_int16 vol = USEX(SCI_VOL);
		if (vol) {
		    vol -= 0x0101;
		    goto setvol;
		}
	    } else if (event == 3) {   /*pause/play -- short press*/
		if (cachedSpiState != spi_pause) {
		    oldSpiState = cachedSpiState;
		    cachedSpiState = spi_pause;

		    /* 'steal' the already read data */
		    stream_save = stream_wr_pointer;
		    stream_wr_pointer = stream_rd_pointer+1;
		} else {
		    cachedSpiState = oldSpiState;
		    /* restore 'stolen' data */
		    stream_wr_pointer = stream_save;
		}
	    } else if (event == -3) {

		if (ui.KeyTime == LONG_PRESS_TIME) { /*pause/play -- long press, one-shot*/
		    /* if pressed 0.75 seconds, toggle pause or random play */
#ifdef NO_RANDOM
		    /* toggle loudness - have LSb set when ON */
		    USEX(SCI_BASS) ^= USEX(SCI_AICTRL2);
		    SetVolumeFromSpi();
#else
		    if (cachedSpiState
#ifdef SW3_RANDOM_PLAY
			!=
#else
			==
#endif
			spi_pause) {
			/* if pause on -- toggle random play */
			USEX(SCI_AICTRL3) ^= CTRL3_RANDOM_PLAY;
		    } else {
			/* toggle loudness - have LSb set when ON */
			USEX(SCI_BASS) ^= USEX(SCI_AICTRL2);
			SetVolumeFromSpi();
		    }
#endif
		} else if (ui.KeyTime == 4*32) {
		    /* if pressed for 4 seconds, power off. */
		    Off();
		}
	    
	    } else if (event == 1 || event == 2) {
		/* short press of next/prev buttons */
		register u_int16 song = USEX(SCI_AICTRL0);
		if (event == 2) {
		    song += 1; /* next file */
		} else if (USEX(SCI_DECODE_TIME) < 5) {
		    if (song == 0) {
			song = USEX(SCI_AICTRL1)-1;
		    } else { /* previous file */
			song -= 1;
		    }
		    /* else the same file */
		}
		USEX(SCI_AICTRL0) = song;
		USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
		cachedSpiState = spi_zeros;
	    }
#endif/*custom/normal */
#else/*3 buttons*/

/*16/32 buttons*/
#ifdef MODE_DIR
/*
  1 - toggle dir (file 0..7 and 8..15)
  2,3 - volume of left chn
  4,5 - volume of right chn
  6-  - select file #-6
 */
	    if (abs(event) == 1) {
#if 1
		USEX(SCI_AICTRL0) ^= 8;
#else /* 1**) for 10 files, see also 2**) */
		if (USEX(SCI_AICTRL0) < 10) {
		    USEX(SCI_AICTRL0) += 10;
		} else {
		    USEX(SCI_AICTRL0) -= 10;
		}
#endif
		goto selSong;
	    } else if (abs(event) == 2) {
/*volume down left -- both short and long presses*/
		register u_int16 vol = USEX(SCI_VOL);
		if ((vol & 0xff00) < 0xff00)
		    vol += 0x0100;
	    setvol:
		USEX(SCI_VOL) = vol;
		SetVolumeFromSpi();
	    } else if (abs(event) == 3) {
/*volume up left -- both short and long presses*/
		register u_int16 vol = USEX(SCI_VOL);
		if ((vol & 0xff00))
		    vol -= 0x0100;
		goto setvol;
	    } else if (abs(event) == 4) {
/*volume down right -- both short and long presses*/
		register u_int16 vol = USEX(SCI_VOL);
		if ((vol & 0x00ff) < 0x00ff)
		    vol += 0x0001;
		goto setvol;
	    } else if (abs(event) == 5) {
/*volume up right -- both short and long presses*/
		register u_int16 vol = USEX(SCI_VOL);
		if ((vol & 0x00ff))
		    vol -= 0x0001;
		goto setvol;
	    } else if (abs(event) >= 6) {
		/* short/long press of button 6..15 to directly select song #*/
#if 1
		USEX(SCI_AICTRL0) = (USEX(SCI_AICTRL0) & 8)
		    + ((abs(event) - 6) & 7);
#else /* 2**) for 10 files, see also 1**) */
		if (USEX(SCI_AICTRL0) < 10) {
		    USEX(SCI_AICTRL0) = (abs(event) - 6);
		} else {
		    USEX(SCI_AICTRL0) = 10 + (abs(event) - 6);
		}
#endif
	    selSong:
		USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
		cachedSpiState = spi_zeros;
		ui.KeyTime = -32768; /*one-shot*/
	    }
#else/*MODE_DIR*/
#ifdef SONG_BUTTONS /*Custom/normal*/
	    if (
#ifdef SHORT_AND_LONG_SELECT_SONG
		abs(event)
#else
		event
#endif
		> 0) {
		/* short/long press of any button to directly select song # */
		USEX(SCI_AICTRL0) = abs(event) -1;
		USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
		cachedSpiState = spi_zeros;
#ifdef SHORT_AND_LONG_SELECT_SONG
		ui.KeyTime = -32768; /*one-shot*/
#endif
	    }

#else/*SONG_BUTTONS*/
	    if (abs(event) == 1) {
/*volume down -- both short and long presses*/
		register u_int16 vol = USEX(SCI_VOL);
		vol += 0x0101;
	    setvol:
		USEX(SCI_VOL) = vol;
		SetVolumeFromSpi();
#ifdef SAVE_VOLUME
		volChanged = 1;
#endif/*SAVE_VOLUME*/
	    } else if (abs(event) == 2) {
/*volume up -- both short and long presses*/
		register u_int16 vol = USEX(SCI_VOL);
		if (vol) {
		    vol -= 0x0101;
		    goto setvol;
		}
	    } else if (event == 3) {
/*pause/play -- short press*/
		if (cachedSpiState != spi_pause) {
		    oldSpiState = cachedSpiState;
		    cachedSpiState = spi_pause;

		    /* 'steal' the already read data */
		    stream_save = stream_wr_pointer;
		    stream_wr_pointer = stream_rd_pointer+1;
		} else {
		    cachedSpiState = oldSpiState;
		    /* restore 'stolen' data */
		    stream_wr_pointer = stream_save;
		}
	    } else if (event == -3) {
/*pause / play -- long press*/
#ifdef NO_RANDOM
		/* toggle loudness - have LSb set when ON */
		USEX(SCI_BASS) ^= USEX(SCI_AICTRL2);
		SetVolumeFromSpi();
#else
		if (cachedSpiState
#ifdef SW3_RANDOM_PLAY
		    !=
#else
		    ==
#endif
		    spi_pause) {
		    /* if pause on -- toggle random play */
		    USEX(SCI_AICTRL3) ^= CTRL3_RANDOM_PLAY;
		} else {
		    /* toggle loudness - have LSb set when ON */
		    USEX(SCI_BASS) ^= USEX(SCI_AICTRL2);
		    SetVolumeFromSpi();
		}
#endif
		ui.KeyTime = -32767; /*makes it one-shot long-press!*/
	    
	    } else if (event == 4 || event == 5) {
		/* long or short press of next/prev buttons */
		register u_int16 song = USEX(SCI_AICTRL0);
		if (event == 5) {
		    song += 1; /* next file */
		} else if (USEX(SCI_DECODE_TIME) < 5) {
		    if (song == 0) {
			song = USEX(SCI_AICTRL1)-1;
		    } else { /* previous file */
			song -= 1;
		    }
		    /* else the same file */
		}
		USEX(SCI_AICTRL0) = song;
		goto songselected;
#ifdef MODE_ABC
#ifdef MODE_ABC_OWN_KEYS
	    } else if (event == 6) {
		/* pause before play */
		USEX(SCI_AICTRL3) =
		    (USEX(SCI_AICTRL3) & ~CTRL3_PLAY_MODE_MASK)
		    | CTRL3_PAUSE_BEFORE_PLAY;
	    } else if (event == 7) {
		/* loop mode */
		USEX(SCI_AICTRL3) =
		    (USEX(SCI_AICTRL3) & ~CTRL3_PLAY_MODE_MASK)
		    | CTRL3_LOOP_SONG;
	    } else if (event == 8) {
		/* normal mode */
		USEX(SCI_AICTRL3) &= ~CTRL3_PLAY_MODE_MASK;
	    } else if (event >= 9) {
		/* short press of any other button
		   to directly select song # */
		USEX(SCI_AICTRL0) = event -9;
	    songselected:
		USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
		cachedSpiState = spi_zeros;
	    } else if (event < 0) {
		/* long presses */
	    }
#else
	    } else if (event == -4) { /* prev, next, song#1 = MODE A, B, C */
		/* pause before play */
		USEX(SCI_AICTRL3) =
		    (USEX(SCI_AICTRL3) & ~CTRL3_PLAY_MODE_MASK)
		    | CTRL3_PAUSE_BEFORE_PLAY;
	    } else if (event == -5) {
		/* loop mode */
		USEX(SCI_AICTRL3) =
		    (USEX(SCI_AICTRL3) & ~CTRL3_PLAY_MODE_MASK)
		    | CTRL3_LOOP_SONG;
	    } else if (event == -6) {
		/* normal mode */
		USEX(SCI_AICTRL3) &= ~CTRL3_PLAY_MODE_MASK;
	    } else if (event > 0) {
		/* short press of any other button
		   to directly select song # */
		USEX(SCI_AICTRL0) = event -6;
	    songselected:
		USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
		cachedSpiState = spi_zeros;
	    } else if (event < 0) {
		/* long presses */
	    }
#endif/*MODE_ABC_OWN_KEYS*/
#else/*MODE_ABC*/

#ifdef SHORT_AND_LONG_SELECT_SONG
	    } else if (abs(event) >= 6) {
		/* short press or long press of any other button
		   to directly select song # */
		USEX(SCI_AICTRL0) = abs(event) -6;
	    songselected:
		USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
		cachedSpiState = spi_zeros;
#ifdef SHORT_AND_LONG_SELECT_SONG
		ui.KeyTime = -32768; /*one-shot*/
#endif
	    }
#else
	    } else if (event > 0) {
		/* short press of any other button
		   to directly select song # */
		USEX(SCI_AICTRL0) = event -6;
	    songselected:
		USEX(SCI_MODE) |= (1<<SCIMB_OUT_OF_WAV);
		cachedSpiState = spi_zeros;
	    } else if (event < 0) {
		/* long presses */
	    }
#endif

#endif/*elseMODE_ABC*/
#endif/*elseSONG_BUTTONS*/
#endif/*elseMODE_DIR*/
#endif/*else3 buttons*/
	}
    }

    if (cachedSpiState == spi_file) {
	/* extra byte read if odd file size */
	spix.size = (fatInfo.fileSize+1)/2; /* in words! */
	cachedSpiState = spi_seek;
setfrag:
	spix.fragment = &fragments[0];
	spix.sector   = fragments[0].start /*& 0x7fffffffUL*/;
	spix.fragSize = fragments[0].size; /* *fatInfo.sectorsPerCluster*/
    }

    if (cachedSpiState == spi_seek) {
#if defined(SAVE_VOLUME)
    /*Save volume when volume button is released. */
	if (ui.Data == 0 &&
	    volChanged
	    //&& cachedSpiState == spi_seek /* already checked for */
	    ) {
	    volChanged = 0;
	    SaveWord(26/*volume=26*/, USEX(SCI_VOL));
	}
#endif/*SAVE_VOLUME*/

#ifdef USE_HC
	MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,
		   (spix.sector & 0x7fffffffUL) << hcShift);
#else
	MmcCommand(MMC_READ_SINGLE_BLOCK|0x40,(spix.sector<<9));
#endif

	spix.sector++;
	if (spix.fragSize-1 == 0) {
	    spix.fragment++;
	    spix.sector  = spix.fragment->start /*& 0x7fffffffUL*/;
	    spix.fragSize = spix.fragment->size;/* *fatInfo.sectorsPerCluster*/
	} else {
	    spix.fragSize = spix.fragSize-1;
	}
	spiCnt = 0;
	cachedSpiState = spi_waitdata;
#ifdef HDATCHECK
	if (++hdatCheck == HDATCHECK && USEX(SCI_HDAT1) == 0) {
	    /* if file not detected at HDATCHECK/2 kB, skip the rest */
	    goto end_of_file;
	}
#endif
    }
    if (cachedSpiState == spi_waitdata) {
	if (SpiSendReceiveMmc(0xff00,8) == 0xfe) {
	    cachedSpiState = spi_data;
	    spiCnt = 0;
#ifdef TIMEOUT
	    timeOut = 0;
#endif
	} else {
	    if ((s_int16)++spiCnt < 0)
		MyReset(); /* timeout */
	}
    } else if (cachedSpiState == spi_data) {
	while (spix.size > 0 &&
	       StreamDiff() < STREAM_BUFFER_SZ-40) {
	    register int i, j = 32;
	    register u_int16 *wr = (void *)stream_wr_pointer;
	    if (j > spix.size)
		j = spix.size;
	    for (i=0;i<j;i++) {
		*wr++ = SpiSendReceiveMmc(-1/*0xffffU*/,16);
		if (wr >= &stream_buffer[STREAM_BUFFER_SZ])
		    wr = stream_buffer;
	    }
	    stream_wr_pointer = (void *)wr;
	    spix.size -= j;//32;
	    spiCnt += 32;
	    if (spiCnt >= 256) {
		/* 256 words read, get next block in file */
		SpiSendReceiveMmc(-1/*0xffff*/,16); /* discard crc */
		cachedSpiState = spi_seek;
		break;
	    }
#ifdef TIMEOUT
	    timeOut = 0;
#endif
	    /* watchdog clearing here doesn't work with midi */
	}

#if defined(TIMEOUT)
	if (
#if !defined(NO_MIDI) /* exclude MIDI from the check */
	    USEX(SCI_HDAT1) != (('M'<<8)|'T') &&
#endif
	    (s_int16)++timeOut == -1) {
	    spix.size = 0;
	}
#endif

#ifdef ENABLE_WATCHDOG
	USEX(WDOG_RESET) = WDOG_RESET_VAL;
#endif
	if (spix.size <= 0) {
end_of_file:
	    /* Discard any unread data, including CRC, before starting again */
	    {
		register int i;
		for (i=0; i<256+2; i++)
		    SpiSendReceiveMmc(-1/*0xffff*/,16);
	    }
	    SpiSendClocks();

#if defined(USE_PLAY_MODES)
	    {
		register int m = (USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK);
		if (m == CTRL3_LOOP_SONG) {
		    /* restart the same file immediately */
		    cachedSpiState = spi_file;
		} else if (m == CTRL3_PAUSE_AFTER_PLAY) {
		    oldSpiState = spi_zeros;
		    cachedSpiState = spi_pause;
		} else {
		    USEX(SCI_AICTRL0) += 1; /* next file */
		    cachedSpiState = spi_zeros;
		}
	    }
#else/*USE_PLAY_MODES*/

#ifdef LOOP_FILES
	    /* restart the same file immediately */
	    cachedSpiState = spi_file;
#else/*LOOP_FILES*/
#ifdef PAUSE_AFTER_PLAY
	    oldSpiState = spi_zeros;
	    cachedSpiState = spi_pause;
#else
	    USEX(SCI_AICTRL0) += 1; /* next file */
	    cachedSpiState = spi_zeros;
#endif
#endif/*else LOOP_FILES*/

#endif/*else USE_PLAY_MODES*/

	}
#ifdef NO_WMA /* reject WMA files */
	if (USEX(SCI_HDAT1) == (('W'<<8)|'M')) {
	    USEX(SCI_AICTRL0) += 1; /* next file */
	    goto goreset;//MyReset();
	}
#endif
    } else if (cachedSpiState == spi_zeros) {
	if (StreamDiff() < STREAM_BUFFER_SZ-40) {
	    register int i, z = 0;
	    register u_int16 *wr = (void *)stream_wr_pointer;
	    for (i=0;i<32;i++) {
		*wr++ = z;
		if (wr >= &stream_buffer[STREAM_BUFFER_SZ])
		    wr = stream_buffer;
		stream_wr_pointer = (void *)wr;
	    }
	    ++spiCnt;
	    /* watchdog clearing here doesn't work with midi */
#ifdef TIMEOUT
	    timeOut = 0;
#endif
	}
#ifdef TIMEOUT
	if ((s_int16)timeOut++ < 0) {
	    goto goreset;
	}
#endif

	if (spiCnt >= 256+2048/32) { /* was left to max 256 by spi_data */
	    /* 2048 words instead of bytes */
goreset:
#if 0
	/* Discard any unread data, including CRC, before starting again */
	{
	    register int i;
	    for (i=0; i<256+2; i++)
		SpiSendReceiveMmc(-1/*0xffff*/,16);
	}
#endif
	    MyReset();
	}
#ifdef ENABLE_WATCHDOG
	goto wdogack;//USEX(WDOG_RESET) = WDOG_RESET_VAL;
#endif
    } else {
#ifdef ENABLE_WATCHDOG
wdogack:
	USEX(WDOG_RESET) = WDOG_RESET_VAL;
#endif
    }

#ifdef USE_LED
    //if (cachedSpiState != spi_pause)
    {
	register int t = 0;
	if ((ui.Timer & 0x18) == 0) {
	    /* DREQ indicates random play status */
#ifdef NO_RANDOM
	    t = USEX(SCI_BASS); /* has lowest bit set if enabled*/
#else
	    t = USEX(SCI_AICTRL3);  /* lowest bit is global random play */
#ifdef LOUDNESS_ALWAYS_ON_OFF
	    /* indicate random play */
#else /*LOUDNESS_ALWAYS_ON_OFF*/

	    if (cachedSpiState
#ifdef SW3_RANDOM_PLAY
		==
#else
		!=
#endif
		spi_pause)
		t = USEX(SCI_BASS); /* has lowest bit set if enabled*/
#endif /*else LOUDNESS_ALWAYS_ON_OFF*/
#endif /*else NO_RANDOM*/
	}

	USEX(SER_DREQ) = t;
    }
#else/*USE_LED*/
    USEX(SER_DREQ) = 0;
#endif/*else USE_LED*/

#if 0
    if (ui.Timer & 0x30)
	USEX(UART_DATA) = 0x00;
#endif
    spiState = cachedSpiState;
#ifdef KEYTIMER
    if (ui.KeyTimer <= 0)
	USEX(UART_DATA) = 0x00;
#endif
}


auto void MyMain(register s_int16 __a1 fullReset) {

    USEX(SCI_MODE) = (USEX(SCI_MODE) & ~(1<<SCIMB_OUT_OF_WAV))
	| ((1<<SCIMB_LINE) | (1<<SCIMB_SDI_NEWMODE));
    USEX(GPIO_ODATA) = SPI_xCS|SPI_xCS2;
    USEX(GPIO_DDR) = SPI_xCS|SPI_CLK|SPI_xCS2;
#if 0 /* set by boot (see c.s) */
    USEX(MODU_DIV) = 32;/*must be small enough -> high enough samplerate!*/
#endif

#ifdef SHORT_AND_LONG_SELECT_SONG
    {
register s_int16 ke = ui.Key;
register s_int16 tm = ui.KeyTime;
#endif

#ifdef NO_RANDOM
    InitHardware(); /* clears memory, so can't be used for debugging */
#else
    {
#ifdef SHUFFLE_PLAY
	register s_int16 r0 = shuffleSeeds[0];
	register s_int16 r1 = shuffleSeeds[1];
#else
	register int r = rand_seed;
#endif
#ifdef KEYTIMER
	register int t = ui.KeyTimer;
#endif
	InitHardware(); /* clears memory, so can't be used for debugging */
#ifdef KEYTIMER
	ui.KeyTimer = t;
	USEX(UART_DIV) = 0x7f7f;//0x3f46; /* 9600 */
#endif
#ifdef SHUFFLE_PLAY
	shuffleSeeds[0] = r0;
	shuffleSeeds[1] = r1;
#else
	rand_seed = r;
#endif
    }
#endif/*else NO_RANDOM*/
#ifdef SHORT_AND_LONG_SELECT_SONG
ui.Key = ke;
ui.KeyTime = tm;
    }
#endif

    InitLayer3();
    InitCommon();
    //InitReaders();
    InitDecode();
    //InitGeneral();
    //AudioInit();
    //if (fullReset)
//    SetHardware(1, 500);
    /* If not powerup reset, re-enable analog immediately. */
    if (!fullReset) {
	USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN2);
    }
#if defined(ENABLE_WATCHDOG) && defined(KEEP_DRIVERS_OFF)
    else {
	/*analog power off and analog driver powerdown */
	USEX(SCI_STATUS) |= (1<<SCIST_APDOWN1)|(1<<SCIST_APDOWN2);
#if LINEIN_BUTTONS == 3
	/* ignore the release of power button */
	ui.Key = 3;
	ui.KeyTime = -32768;
#endif
    }
#else
    else {
#if LINEIN_BUTTONS == 3
	/* ignore the release of power button */
	ui.Key = 3;
	ui.KeyTime = -32768;
#endif
    }
#endif


#if defined(SAVE_POSITION) && !defined(NO_WMA) //&& LINEIN_BUTTONS != 3
    //if (USEX(SCI_AICTRL0))
	SaveWord(28/*AICTRL0*/, 0);
#endif

#ifdef ENABLE_UI
    applAddr = ApplAddr;
#endif
    /* Allow DREQ and enable interrupts, VERY IMPORTANT! */
    //USEX(INT_GLOB_ENA) = 0;
    Enable(); //USEX(INT_GLOB_ENA) = 0;
    InitReadFrame();

    /* wake up MMC! -- If this takes too long, watchdog will bite. */
    {
	register s_int16 i, cmd;
    tryagain:
	for (i=0; i<512; i++) {
	    SpiSendClocks();
	}
	/* MMC Init, command 0x40 should return 0x01 if all is ok. */
	i = MmcCommand(MMC_GO_IDLE_STATE/*CMD0*/|0x40,0);
	if (i != 1)
	    goto tryagain;//continue; /* No valid idle response */
	cmd = MMC_SEND_OP_COND|0x40;
#ifdef USE_HC
	//MmcCommand(MMC_CRC_ON_OFF/*CMD59*/|0x40, 0);
	if (MmcCommand(MMC_SEND_IF_COND/*CMD8*/|0x40, 0x00000122/*2.7-3.6V*/)
	    == 1) {
	    /* MMC answers: 0x05 SD answers: 0x01 */
	    /* Should we read the whole R7 response? */
	    //SpiSendReceiveMmc(-1, 32);
	    cmd = 0x40|41; /* ACMD41 - SD_SEND_OP_COND */
	}
#endif /*USE_HC*/
	/* MMC Wake-up call, set to Not Idle (mmc returns 0x00)*/
	//i = 0; /* i is 1 when entered .. but does not make the code shorter*/
	while (1) {
	    if (MmcCommand(cmd/*MMC_SEND_OP_COND|0x40*/,
#ifdef USE_HC
			   /* Support HC for SD, but not for MMC! */
			   cmd == (0x40|41) ?
			   MmcCommand(55|0x40/*CMD55*/,0), 0x40000000UL : 0
#else
			   0
#endif
		) == 0)
		break;
	    if (++i < 0 /*> 32000*/) {
		goto tryagain; /* Not able to power up mmc */
	    }
	}

#ifdef USE_HC
	hcShift = 9;
	i = 0x0246;
	/* PRETEC's 1G MMC does not seemd to handle CMD58! */
	/* Support HC for SD, but not for MMC! */
	if (cmd == (0x40|41) &&
	    MmcCommand(MMC_READ_OCR/*CMD58*/|0x40, 0) == 0) {
	    if (SpiSendReceiveMmc(-1, 16) & (1<<(30-16))) {
		/* OCR[30]:CCS - card capacity select */
		/* HighCapacity! */
		/* ldc 9,D0 (0x0246) <-> ldc 0,D0 (0x0006) */
		i = 6;
		hcShift = 0;
	    }
	    //SpiSendReceiveMmc(-1, 16);
	}
	WriteToProgramRam((u_int16)ReadDiskSector+2, 0, i);
#endif /*USE_HC*/

	/* Set Block Size of 512 bytes -- default for at least HC */
	/* Needed by MaxNova S043618ATA 2J310700 MV016Q-MMC */
	/* Must check return value! (MicroSD restart!) */
	if (MmcCommand(MMC_SET_BLOCKLEN|0x40, 512) != 0)
	    goto tryagain;
	/* All OK return */
    }

    /* Detects cards also without partition table */
    if (InitFileSystem() == 0 && (USEX(SCI_AICTRL1) = OpenFile(0xffffU))) {
	/* determine num of files */

#if !defined(NO_RANDOM) /* if random play feature active */
	if ((USEX(SCI_AICTRL3) & CTRL3_RANDOM_PLAY)) {
#ifdef SHUFFLE_PLAY /*shuffle 23 words*/
	    USEX(SCI_AICTRL0) = Shuffle(USEX(SCI_AICTRL1), USEX(SCI_AICTRL0));
#else
	    /* random play (not shuffle!) */
	    USEX(SCI_AICTRL0) = (myrand() % USEX(SCI_AICTRL1));
#endif
	}
#endif

	if (OpenFile(USEX(SCI_AICTRL0)) < 0) {
#if defined(SKIP_ID3V2) /* 48 words */
	    /* check for ID3v2 to speed up start of playback */
	    /* An ID3v2 tag can be detected with the following pattern:
	       $49 44 33 yy yy xx zz zz zz zz */
	    ReadDiskSector(fragments[0].start);
	    if ((GetLong(0) & 0xffffffL) == 0x334449) {
		/* skip in 16kB resolution, assume tags are smaller than 2MB */
#if 1
		register u_int16 skip = GetByte(7) << 5;
#else
		/* 0.5kB resolution now */
		register u_int16 skip = 32 * GetByte(7) + GetByte(8)/4;
#endif
		/* guard against small fragments */
		if (skip >= fragments[0].size)
		    skip = fragments[0].size-1;
		fragments[0].start += skip;
		fragments[0].size  -= skip;
		fatInfo.fileSize -= (u_int32)skip * 512;
	    }
#endif

#if defined(USE_PLAY_MODES)
	    if ((USEX(SCI_AICTRL3) & CTRL3_PLAY_MODE_MASK) ==
		CTRL3_PAUSE_BEFORE_PLAY) {
#ifdef PAUSE_BEFORE_TO_AFTER
		USEX(SCI_AICTRL3) = CTRL3_PAUSE_AFTER_PLAY;
#endif
		oldSpiState = spi_file;
		spiState = spi_pause;
		stream_save = stream_wr_pointer;
	    } else {
		spiState = spi_file;
	    }
#else /*USE_PLAY_MODES*/

#ifdef PAUSE_BEFORE_PLAY
	    /* Mark file ready and go to pause state in UserHook */
	    oldSpiState = spi_file;
	    spiState = spi_pause;
	    stream_save = stream_wr_pointer;
#else
	    spiState = spi_file;
#endif
#endif /*else USE_PLAY_MODES*/

#ifdef HDATCHECK
	    hdatCheck = 0;
#endif
	} else {
	gofirst:
	    USEX(SCI_AICTRL0) = 0; /* file not found, jump to first file */
	    MyReset();
	}
    } else {
	/* Initialize again */
#if LINEIN_BUTTONS == 3
	/* No FAT or no files, power off */
	Off();
#else
	goto gofirst;//MyReset();
#endif
    }

#if defined(SAVE_POSITION) //&& LINEIN_BUTTONS != 3
    SaveWord(28/*AICTRL0*/, USEX(SCI_AICTRL0));
#endif

#if defined(ENABLE_WATCHDOG) && defined(KEEP_DRIVERS_OFF)
    /*otherwise keep it in powerdown until MMC found*/
    USEX(SCI_STATUS) &= ~(1<<SCIST_APDOWN1); /* enable ANALOG power */
#endif
    timeToRemovePDown2 = 47;

    USEX(INT_ENABLE) = (1<<INT_EN_SCI)|(1<<INT_EN_DAC)|(1<<INT_EN_RX);
    USEX(SCI_DECODE_TIME) = 0;
    while (1) {
	if (!ReadFrame()) {
	    play_frame();
	} else {
	    InitReadFrame();
	}
    }
}
